#include "Database.h"
#include "Image.h"

#include <Base\Containers\range_for_std_containers.hpp>
#include <Base\IO\FileUtil.hpp>
#include <Base\IO\MemoryReadStream.h>
#include <Base\IO\ReadStreamBinLex.h>
#include <Base\IO\ReadStreamBinLex_stdvector.h>
#include <Base\IO\ReadStreamZLib.h>
#include <Base\String\Formatting.h>

namespace hyplex::drakan
{
	static const bool kEnableDump = true;

	/*


	Texture decode read at 0041D75E.

	*/

	ENodeType Texture::getType() const noexcept
	{
		return ENodeType::TEXTURE;
	}

	void Texture::consumeData(const Entry& entry, BasicBufferedLexState& bin)
	{
		if (entry.type == 0x40)
		{
#pragma pack(push)
#pragma pack(1)
			const struct Header
			{
				uint32_t width = 0;
				uint32_t height = 0;
				uint32_t pitch = 0;
				uint16_t bitsPerPixel = 0;
				uint32_t alphaBitsPerPixel = 0;
				uint32_t unk5 = 0;
				uint32_t unk6 = 0;
				uint32_t unk7 = 0;
				uint32_t bumpMapTexId = 0;
				uint32_t unk9 = 0;
				uint32_t materialObjectId = 0;
				uint32_t unk11 = 0;
				uint32_t unk12 = 0;
				uint32_t compressedSize = 0;
			} header = read<Header>(bin);
#pragma pack(pop)

			id = entry.unk00;
			width = header.width;
			height = header.height;
			pitch = header.pitch;
			bitsPerPixel = header.bitsPerPixel;
			alphaBitsPerPixel = header.alphaBitsPerPixel;
			unk5 = header.unk5;
			unk6 = header.unk6;
			unk7 = header.unk7;
			bumpMapTexId = header.bumpMapTexId;
			animationStartFrameFPS = header.unk9 & 0xFF;
			isAnimationNextFrame = (header.unk9 & 0x2000) != 0;
			materialObjectId = header.materialObjectId;
			unk11 = header.unk11;
			unk12 = header.unk12;

			ZLibBytesStreamFromChunksStream<MemoryReadStream> zlibStream(MemoryReadStream(consumeUptoN(bin, header.compressedSize)));
			blob.resize(pitch * height);
			HYPLEX_ENFORCE(zlibStream.read(blob) == blob.size());
			HYPLEX_ENFORCE(zlibStream.read(blob) == 0);

			
			/*if (header.alphaBitsPerPixel > 0)
			{
				const unsigned int alphaMapCompressedSize = read<uint32_t>(bin) * header.alphaBitsPerPixel / 8;
				zlibStream = MemoryReadStream(consumeUptoN(bin, alphaMapCompressedSize));
				HYPLEX_ENFORCE(zlibStream.read(blob) == blob.size());
			}*/

			// Has unknown stuff at the end.
			//if(entry.unk00 != 0x1622)
			//	HYPLEX_ENFORCE(!peek(bin));
		}
	}

	static std::array<uint8_t, 4> decode12RGB4A(uint16_t foo)
	{
		/*return {
			uint8_t((uint8_t(foo >> 8) & 0b1111) * 0b00010001),
			uint8_t((uint8_t(foo >> 4) & 0b1111) * 0b00010001),
			uint8_t((uint8_t(foo) & 0b1111) * 0b00010001),
			uint8_t((uint8_t(foo >> 12) & 0b1111) * 0b00010001)
		};*/
		return {
			uint8_t((uint8_t(foo >> 8) & 0b1111) * 0b00010000),
			uint8_t((uint8_t(foo >> 4) & 0b1111) * 0b00010000),
			uint8_t((uint8_t(foo) & 0b1111) * 0b00010000),
			uint8_t((uint8_t(foo >> 12) & 0b1111) * 0b00010000)
		};
	}

	static std::array<uint8_t, 4> decode15RGB1A(uint16_t foo)
	{
		return {
			uint8_t((uint8_t(foo >> 10) & 0b11111) * 0b00001000),
			uint8_t((uint8_t(foo >> 5) & 0b11111) * 0b00001000),
			uint8_t((uint8_t(foo) & 0b11111) * 0b00001000),
			uint8_t(foo >> 15 ? 0xFF : 0)
		};
	}

	static std::array<uint8_t, 4> decode16RGB(uint16_t foo)
	{
		return {
			uint8_t(((foo >> (5+6)) & 0b11111) * 0b00001000),
			uint8_t(((foo >> 5) & 0b111111) * 0b00000100),
			uint8_t((foo & 0b11111) * 0b00001000),
			255
		};
	}

	void Texture::dumpNode(const NodeDumpArg& arg) const
	{
		if (!kEnableDump)
			return;

		std::vector<std::array<uint8_t, 4>> rgba;
		rgba.resize(width * height);

		if (bitsPerPixel == 16)
		{
			for (unsigned int y = 0; y < height; ++y)
			{
				for (unsigned int x = 0; x < width; ++x)
				{
					const uint8_t low = blob[y * pitch + x * 2];
					const uint8_t high = blob[y * pitch + x * 2 + 1];
					const uint16_t foo = low | (high << 8);

					auto& pixel = rgba[(height - 1 - y) * width + x];
					pixel = alphaBitsPerPixel == 4 ? decode12RGB4A(foo) :
						alphaBitsPerPixel == 1 ? decode15RGB1A(foo) :
						decode16RGB(foo);
				}
			}
		}
		else if (bitsPerPixel == 8)
		{
			const auto palette = asArrayView(arg.db->palette);
			for (unsigned int y = 0; y < height; ++y)
			{
				for (unsigned int x = 0; x < width; ++x)
				{
					const uint8_t foo = blob[y * pitch + x];

					auto& pixel = rgba[(height - 1 - y) * width + x];
					pixel = foo < palette.size() ? palette[foo] : std::array<uint8_t, 4>{ 255, 0, 255, 255 };
					pixel[3] = 255;
				}
			}
		}


		struct : gfx::IImage
		{
			uivec2 dimensions;
			RawConstMemoryRange blob;

			virtual gfx::EPixelFormat getPixelFormat() const override
			{
				return gfx::EPixelFormat::U8RGBA;
			}
			virtual uivec2 getDimensions() const override
			{
				return dimensions;
			}
			virtual unsigned int getBytesPerRow() const override
			{
				return dimensions.x * 4;
			}
			virtual unsigned int getNumMipMaps() const override
			{
				return 1;
			}
			virtual RawConstMemoryRange getRawData() const override
			{
				return asRawUChars(asArrayView(blob));
			}
		} img;

		img.dimensions = { width, height };
		img.blob = asRawUChars(asArrayView(rgba));

		const std::vector<char> pngBlob = gfx::makePNG(img);

		FileUnbufferedWriteStream(sprintfmt("%sTexture%u.png", arg.dbRootDir, id), EFileWriteOpenMode::RECREATE, true)
			//.write(asRawUChars(asArrayView(blob)));
			.write(asRawUChars(asArrayView(pngBlob)));

		/*if (id == 5673)
		{
			FileUnbufferedWriteStream(sprintfmt("%s%u.texture.bin", dir, id), EFileWriteOpenMode::RECREATE, true)
				.write(asRawUChars(asArrayView(blob)));
			FileUnbufferedWriteStream(sprintfmt("%s%u.texture.rgba", dir, id), EFileWriteOpenMode::RECREATE, true)
				.write(asRawUChars(asArrayView(img.blob)));
		}*/
	}

	void Texture::dumpAnimationInfo(const Directory& textureDirectory, string_view outputDir)
	{
		(void)outputDir;

		String str;

		std::vector<const Texture*> animTextureList;

		const auto flushAnim = [&] {
			if (animTextureList.size() >= 2)
			{
				printfmt(str, "%u fps", animTextureList[0]->animationStartFrameFPS);
				for (const Texture* const frame : animTextureList)
					printfmt(str, " %u", frame->id);
				str += '\n';
			}
		};

		for (const auto& node : textureDirectory.children)
		{
			if (node->getType() != ENodeType::TEXTURE)
				continue;

			const Texture* const texture = static_cast<const Texture*>(node.get());

			if (texture->isAnimationNextFrame)
			{
				// Fails for Ever Tear\data01.
				//HYPLEX_ENFORCE(!animTextureList.empty());
			}
			else
			{
				flushAnim();
				animTextureList.clear();
			}

			animTextureList.push_back(texture);
		}

		flushAnim();

		if (!str.empty())
		{
			FileUnbufferedWriteStream(sprintfmt("%sTexture Animation Data.txt", outputDir), EFileWriteOpenMode::RECREATE, true)
				.write(asRawUChars(asArrayView(str)));
		}
	}
}